# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from typing import Any
from collections.abc import Callable

from concurrent.futures import ThreadPoolExecutor, Future
from concurrent.futures import wait as wait_futures


class SyncExecutor(ThreadPoolExecutor):
    def __init__(self, max_workers: int | None = None) -> None:
        super().__init__(max_workers)
        self.futures: list[Future[Any]] = []

    def submit(self, fn: Callable[..., Any], /,
               *args: Any, **kwargs: Any) -> Future[Any]:
        future = super().submit(fn, *args, **kwargs)
        self.futures.append(future)
        return future

    def shutdown(self, wait: bool = True, *,
                 cancel_futures: bool = False) -> None:
        super().shutdown(wait, cancel_futures=cancel_futures)

    def raise_exceptions(self, timeout: int | None = None) -> None:
        done, not_done = wait_futures(self.futures, timeout)

        for future in done:
            future.result()



"""Some UI and OS-dependent utility functions."""

# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import os
import sys
import platform
import subprocess
from pathlib import Path
from enum import IntEnum
from datetime import datetime, timezone

from PyQt6.QtCore import QSettings, QObject


def open_in_file_browser(file: Path) -> None:
    """Open the given file in the system file browser."""

    if sys.platform == "win32":
        os.startfile(file)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", file])
    else:
        subprocess.Popen(["xdg-open", file])


def add_to_startup(app_id: str) -> None:
    """Add the app to start up on macOS."""

    if platform.system() != "Darwin":
        return

    # Set the paths and filenames
    app_path = f"/Applications/{app_id}.app"
    launch_dir = Path("~/Library/LaunchAgents").expanduser()

    if not launch_dir.exists():
        launch_dir.mkdir()

    plist_path = launch_dir / f"{app_id}.plist"
    plist_path.touch()

    # Create the QSettings object
    settings = QSettings(str(plist_path), QSettings.Format.NativeFormat)

    # Set the launch agent properties
    settings.setValue('Label', app_id)
    settings.setValue('ProgramArguments', app_path)
    settings.setValue('RunAtLoad', True)
    settings.setValue('KeepAlive', False)

    # Save the settings to create the plist file
    settings.sync()


class Time(IntEnum):
    SECOND = 1
    MINUTE = 60
    HOUR = MINUTE * 60
    DAY = HOUR * 24
    WEEK = DAY * 7
    MONTH = DAY * 30
    YEAR = DAY * 365


class TimeStrings(QObject):
    def get_name(self, unit: Time, plural: bool) -> str:
        # it's not as simple as appending an 's'
        if plural:
            return self.get_plural_name(unit)

        match unit:
            case Time.SECOND:
                return self.tr("second")
            case Time.MINUTE:
                return self.tr("minute")
            case Time.HOUR:
                return self.tr("hour")
            case Time.DAY:
                return self.tr("day")
            case Time.WEEK:
                return self.tr("week")
            case Time.MONTH:
                return self.tr("month")
            case Time.YEAR:
                return self.tr("year")

    def get_plural_name(self, unit: Time) -> str:
        match unit:
            case Time.SECOND:
                return self.tr("seconds")
            case Time.MINUTE:
                return self.tr("minutes")
            case Time.HOUR:
                return self.tr("hours")
            case Time.DAY:
                return self.tr("days")
            case Time.WEEK:
                return self.tr("weeks")
            case Time.MONTH:
                return self.tr("months")
            case Time.YEAR:
                return self.tr("years")

    @property
    def template(self) -> str:
        return self.tr("{n} {name} ago")


def time_ago(timestamp: datetime) -> str:
    delta = datetime.now(tz=timezone.utc) - timestamp
    s = int(delta.total_seconds())

    def get_human_time(seconds: int, unit: Time) -> str:
        n = seconds // unit
        t = TimeStrings()
        name = t.get_name(unit, n != 1)
        return t.template.format(n=n, name=name)

    previous = Time.SECOND

    for unit in Time:
        if s < unit.value:
            return get_human_time(s, previous)
        previous = unit

    return get_human_time(s, Time.YEAR)


def windows_safe_path(path: Path) -> Path:
    UNC_PREFIX = u'\\\\?\\'

    if platform.system() == "Windows":
        if not path.drive.startswith(UNC_PREFIX):
            return Path(UNC_PREFIX + str(path))
    return path



from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBCourseContent

from .base import FStream
from .job import DownloadJob
from .templates import create_body
from .webdav import WebDavFile, ContentParser


class ContentBody(FStream):
    """Process the content body to find WebDav files."""

    def __init__(self, content: BBCourseContent, _: None,
                 job: DownloadJob) -> None:
        self.ignore = False

        if not content.body:
            self.ignore = True
            return

        title = content.title or "Untitled"
        parser = ContentParser(content.body, job.session.instance_url)

        self.body = create_body(title, parser.body, parser.text)
        self.children = [WebDavFile(ln, job) for ln in parser.links]

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        if self.ignore:
            return

        self.write_base(path / f"{path.stem}.html", executor, self.body)

        for child in self.children:
            child.write(path, executor)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

from importlib.metadata import version as get_version
from importlib.metadata import PackageNotFoundError

from blackboard_sync import __about__

# -- Project information -----------------------------------------------------

project = "Blackboard Sync"
copyright = __about__.__copyright__
author = __about__.__author__

# The full version, including alpha/beta/rc tags
release = get_version("blackboardsync")
version = ".".join(release.split('.')[:2])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'recommonmark',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.doctest',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

autoclass_content = 'both'
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
    '.txt': 'markdown'
}


intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'hypothesis': ('https://hypothesis.readthedocs.io/en/latest', None),
    # 'pathlib': ('https://docs.python.org/3/library/pathlib.html', None)
}

# Make sure the target is unique
autosectionlabel_prefix_document = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'show_powered_by': False,
    'github_user': 'sanjacob',
    'github_repo': 'BlackboardSync',
    'github_banner': True,
    'show_related': False,
    'note_bg': '#FFF59C',
    'github_button': True,
    'github_type': 'star',
    'description': __about__.__summary__,
    # 'analytics_id': 'UA-109078714-2',
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



"""
Checks for updates in GitHub.
"""

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import requests
from packaging import version
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_version


def check_for_updates() -> bool:
    """Checks if there is a newer release than the current on Github."""
    package = __package__.replace('_', '')

    try:
        __version__ = get_version(package)
    except PackageNotFoundError:
        return False

    domain = "https://api.github.com"
    url = f"{domain}/repos/sanjacob/BlackboardSync/releases/latest"

    response = requests.get(url, timeout=2000)

    if response.status_code == 200:
        json_response = response.json()
        tag = json_response['tag_name']
        tag = tag[1:] if tag[0] == 'v' else tag

        return version.parse(tag) > version.parse(__version__)
    return False



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from enum import IntEnum
from pathlib import Path
# from typing import override
from datetime import datetime

from PyQt6.QtCore import pyqtSlot, Qt
from PyQt6.QtWidgets import (QWizard, QWizardPage,
                             QCompleter, QComboBox,
                             QCheckBox, QSpinBox,
                             QLabel, QPushButton)

from .assets import load_ui, get_wizard_pixmap
from .dialogs import DirDialog, Dialogs


class SetupWizard(QWizard):
    """Guide user through initial setup process."""

    class Pages(IntEnum):
        INTRO = 0
        INSTITUTION = 1
        DOWNLOAD_LOCATION = 2
        DOWNLOAD_SINCE = 3
        LAST = 3

    def __init__(self, support_url: str, institutions: list[str],
                 selected: int | None = None):
        super().__init__()

        # Typing information
        self.uni_selection_page: QWizardPage
        self.sync_location_page: QWizardPage
        self.uni_selection_box: QComboBox
        self.since_all_checkbox: QCheckBox
        self.date_spinbox: QSpinBox
        self.autodetect_label: QLabel
        self.sync_location_button: QPushButton

        self._institutions = institutions
        self._support_url = support_url
        self._has_chosen_location = False
        self._init_ui(selected)

    def _init_ui(self, selected: int | None) -> None:
        load_ui(self)

        pixmap_roles = [
            QWizard.WizardPixmap.BackgroundPixmap
        ]

        for role in pixmap_roles:
            self.setPixmap(role, get_wizard_pixmap(role))

        # Institution
        self.uni_selection_box.addItems(self._institutions)
        self.uni_selection_box.clearEditText()
        self.autodetect_label.setVisible(selected is not None)

        self.completer = QCompleter(self._institutions, self)
        self.completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.completer.setFilterMode(Qt.MatchFlag.MatchContains)
        self.uni_selection_box.setCompleter(self.completer)

        # Download location
        self.sync_location_button.clicked.connect(
            self._choose_location
        )

        # Minimum year
        year = datetime.today().year
        self.date_spinbox.setRange(2000, year)
        self.date_spinbox.setValue(year)
        self.date_spinbox.setEnabled(False)
        self.since_all_checkbox.stateChanged.connect(self._toggle_all_content)

        # Prepare wizard
        self.register_fields()

        # Select autodetected university
        # Must happen after the fields are registered
        if selected is not None:
            self.uni_selection_box.setCurrentIndex(selected)

        self.dialogs = Dialogs()

    def register_fields(self) -> None:
        # Make these actions obligatory
        self.uni_selection_page.registerField(
            "userInstitution*",
            self.uni_selection_box.lineEdit()
        )

        self.sync_location_page.registerField(
            "syncLocation*",
            self.sync_location_button,
            property="text",
            changedSignal=self.sync_location_button.clicked
        )

    def _update_download_location_button(self) -> None:
        dir = self.download_location.name or str(self.download_location)
        self.sync_location_button.setText(dir)

    # @override
    def initializePage(self, id: int) -> None:
        if id == self.Pages.DOWNLOAD_LOCATION:
            if self._has_chosen_location:
                self._update_download_location_button()

    # @override
    def validateCurrentPage(self) -> bool:
        valid = True

        # Do not progress if institution is not valid
        if self.currentId() == self.Pages.INSTITUTION:
            if not self._institution_is_valid():
                self.dialogs.uni_not_supported_dialog(self._support_url)
                valid = False

        return valid

    @pyqtSlot()
    def _choose_location(self) -> None:
        location = DirDialog().choose()

        if location is not None:
            self.download_location = location

    @pyqtSlot(int)
    def _toggle_all_content(self, state: int) -> None:
        unchecked = Qt.CheckState(state) != Qt.CheckState.Checked
        self.date_spinbox.setEnabled(unchecked)

    def _institution_is_valid(self) -> bool:
        return self.institution == str(self.field("userInstitution"))

    @property
    def institution(self) -> str:
        """Name of selected institution."""
        return self.uni_selection_box.itemText(self.institution_index)

    @property
    def institution_index(self) -> int:
        """Index of selected institution."""
        return self.uni_selection_box.currentIndex()

    @property
    def download_location(self) -> Path:
        """Download location selected by user."""
        return self._download_location

    @download_location.setter
    def download_location(self, location: Path) -> None:
        self._download_location = location.resolve()
        self._update_download_location_button()
        self._has_chosen_location = True

    @property
    def min_year(self) -> int | None:
        """Minimum year selected by user."""
        if not self.since_all_checkbox.isChecked():
            return self.date_spinbox.value()
        return None



#!/usr/bin/env python3

"""
BlackboardDownload,
mass download all user content from Blackboard
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import logging
from pathlib import Path
from datetime import datetime, timezone

from blackboard.api_extended import BlackboardExtended
from blackboard.filters import BBMembershipFilter, BWFilter

from .executor import SyncExecutor
from .content.job import DownloadJob
from .content.course import Course


logger = logging.getLogger(__name__)


class BlackboardDownload:
    """Blackboard download job."""

    _last_downloaded = datetime.fromtimestamp(0, tz=timezone.utc)

    def __init__(self, sess: BlackboardExtended,
                 download_location: Path,
                 last_downloaded: datetime | None = None,
                 min_year: int | None = None):
        """BlackboardDownload constructor

        Download all files in blackboard recursively to download_location,
        only if they have been altered since specified datetime

        Keyword arguments:

        :param BlackboardExtended sess: UCLan BB user session
        :param (str / Path) download_location: Where files will be stored
        :param str last_downloaded: Files modified before are ignored
        :param min_year: Courses created before are ignored
        """

        self._sess = sess
        self._user_id = sess.user_id
        self._download_location = download_location
        self._min_year = min_year
        self.executor = SyncExecutor()
        self.cancelled = False

        if last_downloaded is not None:
            self._last_downloaded = last_downloaded

    def download(self) -> datetime | None:
        """Retrieve the user's courses, and start download of all contents

        :return: Datetime when method was called.
        """
        if self.cancelled:
            return None

        logger.info("Starting Blackboard content download")

        start_time = datetime.now(timezone.utc)

        if not self.download_location.exists():
            self.download_location.mkdir(parents=True)
            logger.info("Created download folder")

        logger.info("Fetching user memberships and courses")

        course_filter = BBMembershipFilter(min_year=self._min_year,
                                           data_sources=BWFilter())
        courses = self._sess.ex_fetch_courses(user_id=self.user_id,
                                              result_filter=course_filter)

        job = DownloadJob(session=self._sess,
                          last_downloaded=self._last_downloaded)

        for course in courses:
            if self.cancelled:
                break

            logger.info(f"Fetching user course <{course.id}>")

            Course(course, job).write(self.download_location, self.executor)

        logger.info("Shutting down download workers")

        self.executor.shutdown(wait=True, cancel_futures=self.cancelled)
        self.executor.raise_exceptions()

        return start_time if not self.cancelled else None

    def cancel(self) -> None:
        """Cancel the download job."""
        self.cancelled = True

    @property
    def download_location(self) -> Path:
        """The location where files will be downloaded to."""
        return self._download_location

    @property
    def user_id(self) -> str:
        """User ID used for API calls."""
        return self._user_id



from datetime import datetime
from git import Repo

from packaging.version import Version


def get_releases():
    repo = Repo('.')
    releases = []

    for tag in repo.tags:
        ver = tag.path.split("/")[-1]

        if "-" in ver:
            continue

        t = tag.commit.committed_date
        dt = datetime.utcfromtimestamp(t).date()

        releases.append({
            'version': ver,
            'date': dt.isoformat()
        })

    releases.sort(key=lambda x: Version(x['version']), reverse=True)
    return releases



import platform
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBCourseContent

from .base import FStream
from .job import DownloadJob


class ExternalLink(FStream):
    """Creates a platform-aware internet shortcut."""

    def __init__(self, content: BBCourseContent, _: None,
                 job: DownloadJob) -> None:
        self.url = None

        if content.contentHandler is not None:
            self.url = content.contentHandler.url

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        if self.url is None:
            return

        path = path / path.stem

        if platform.system() in ["Windows", "Darwin"]:
            body = f"[InternetShortcut]\nURL={self.url}"
            path = path.with_suffix(".url")
        else:
            body = self.create_unix_body(self.url)

        super().write_base(path, executor, body)

    def create_unix_body(self, url: str) -> str:
        return f"[Desktop Entry]\nIcon=text-html\nType=Link\nURL[$e]={url}"

    @property
    def create_dir(self) -> bool:
        return True



"""IP and RDAP utilities to enable university detection."""

# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import whoisit
import requests

IP_API = "https://api.ipify.org"
IP_TIMEOUT = 4


def find_my_ip() -> str | None:
    try:
        r = requests.get(IP_API, timeout=IP_TIMEOUT)
        r.raise_for_status()
    except requests.RequestException:
        return None
    else:
        return r.text


def find_ip_entity(ip: str):
    try:
        whoisit.bootstrap()
        r = whoisit.ip(ip)
    except whoisit.errors.QueryError:
        return None
    except whoisit.errors.BootstrapError:
        return None
    else:
        return r



"""
Parse an html to look for links
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.


import mimetypes
from pathlib import Path
from requests import Response
from bs4 import BeautifulSoup
from urllib.parse import unquote
from typing import List, NamedTuple
from pathvalidate import sanitize_filename
from concurrent.futures import ThreadPoolExecutor

from .base import BStream
from .job import DownloadJob


class Link(NamedTuple):
    href: str
    text: str


class ContentParser:
    def __init__(self, body: str, base_url: str,
                 *, find_links: bool = True) -> None:
        soup = BeautifulSoup(body, 'html.parser')
        self._links = []

        if find_links:
            a = self._find_replace(soup, 'a', 'href', base_url)
            img = self._find_replace(soup, 'img', 'src', base_url)
            self._links = [*a, *img]

        self._body = str(soup)
        self._text = soup.text

    def _find_replace(self, soup: BeautifulSoup,
                      tag: str, attr: str, base_url: str) -> list[Link]:
        links = []

        for el in soup.find_all(tag):
            # Add link for later download
            uri = el.get(attr)

            if uri:
                # Handle url-encoding
                filename = unquote(uri.split('/')[-1])
                links.append(Link(href=uri, text=filename))

                # Replace for local instance
                if uri.startswith(base_url):
                    el[attr] = filename
        return links

    @property
    def links(self) -> List[Link]:
        return self._links

    @property
    def body(self) -> str:
        return self._body

    @property
    def text(self) -> str:
        return self._text


def validate_webdav_response(response: Response,
                             link: str, base_url: str) -> bool:
    if response.status_code == 200:
        h = response.headers
        content_type = h.get('Content-Type', '')
        content_len = int(h.get('Content-Length', 0))

        # TODO: feature: select mime types
        len_limit = 1024 * 1024 * 20  # 20 MB

        filters = [
            link.startswith(base_url),
            'video' not in content_type,
            content_len < len_limit
        ]

        return all(filters)
    return False


class WebDavFile(BStream):
    """A Blackboard WebDav file which can be downloaded directly"""
    def __init__(self, link: Link, job: DownloadJob) -> None:
        self.title = sanitize_filename(link.text, replacement_text="_")
        self.stream = job.session.download_webdav(webdav_url=link.href)
        content_type = self.stream.headers.get('Content-Type', 'text/plain')
        self.extension = mimetypes.guess_extension(content_type)
        self.valid = validate_webdav_response(self.stream, link.href,
                                              job.session.instance_url)

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        if self.valid:
            path = Path(path, self.title)

            if self.extension:
                path = path.with_suffix(self.extension)

            super().write_base(path, executor, self.stream)



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from enum import IntEnum
from pathlib import Path

from PyQt6.QtCore import pyqtSlot, pyqtSignal, QObject
from PyQt6.QtWidgets import QWidget, QLabel
from PyQt6.QtWidgets import QComboBox, QPushButton, QDialogButtonBox

from .assets import load_ui
from .dialogs import DirDialog


class SettingsWindow(QWidget):
    """Tweaks main application preferences."""

    class SyncPeriod(IntEnum):
        HALF_HOUR = 60 * 30
        ONE_HOUR = 60 * 60
        SIX_HOURS = 60 * 60 * 6

    class Signals(QObject):
        log_out = pyqtSignal()
        setup_wiz = pyqtSignal()
        save = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        # Typing information
        self.frequency_combo: QComboBox
        self.current_session_label: QLabel
        self.download_location_hint: QLabel
        self.version_label: QLabel
        self.select_download_location: QPushButton
        self.log_out_button: QPushButton
        self.setup_button: QPushButton
        self.button_box: QDialogButtonBox

        self.signals = self.Signals()

        self._init_ui()

    def _init_ui(self) -> None:
        load_ui(self)

        # Slots
        self.select_download_location.clicked.connect(self._choose_location)

        # Signals
        self.log_out_button.clicked.connect(self.signals.log_out)
        self.setup_button.clicked.connect(self.signals.setup_wiz)
        self.button_box.accepted.connect(self.signals.save)

    @pyqtSlot()
    def _choose_location(self) -> None:
        location = DirDialog().choose()

        if location is not None:
            self.download_location = location

    @property
    def download_location(self) -> Path:
        """Current download location."""
        return self._download_location

    @download_location.setter
    def download_location(self, location: Path) -> None:
        self._download_location = location.resolve()
        self.download_location_hint.setText(str(self._download_location))

    @property
    def sync_frequency(self) -> int:
        """Seconds to wait between each sync job."""
        options = list(self.SyncPeriod)
        return int(options[self.frequency_combo.currentIndex()])

    @sync_frequency.setter
    def sync_frequency(self, s: int) -> None:
        options = list(self.SyncPeriod)
        current = options.index(self.SyncPeriod(s))
        self.frequency_combo.setCurrentIndex(current)

    @property
    def username(self) -> str:
        """Username of current session."""
        return self.current_session_label.text()

    @username.setter
    def username(self, username: str) -> None:
        if username:
            self.current_session_label.setText(
                self.tr("Logged in as ") + username)
        else:
            self.current_session_label.setText(
                self.tr("Not currently logged in"))

    @property
    def version(self) -> str | None:
        return self.version_label.text()

    @version.setter
    def version(self, value: str | None) -> None:
        if value is None:
            value = self.tr("No version detected")
        self.version_label.setText(value)



#!/usr/bin/env python3

"""
BlackboardSync
~~~~~~~~~~

Keep your Blackboard files synced locally

:copyright: (c) 2023, Jacob Sánchez Pérez.
:license: GPL v2.0, see LICENSE for details.
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
# Also available at https://www.gnu.org/licenses/old-licenses/gpl-2.0.html

import logging

from .__about__ import (
    __id__,
    __title__,
    __summary__,
    __uri__,
    __homepage__,
    __author__,
    __email__,
    __publisher__,
    __license__,
    __license_spdx__,
    __copyright__
)

__all__ = [
    '__id__', '__title__', '__summary__', '__uri__', '__homepage__',
    '__author__', '__email__', '__publisher__', '__license__',
    '__license_spdx__', '__copyright__'
]

# Console Output
logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from typing import NamedTuple
from enum import Enum, IntEnum, auto

from PyQt6.QtWidgets import QSystemTrayIcon
from PyQt6.QtCore import QObject


class Event(Enum):
    UPDATE_AVAILABLE = auto()
    DOWNLOAD_ERROR = auto()
    APP_RUNNING = auto()


class Severity(Enum):
    NO_ICON = QSystemTrayIcon.MessageIcon.NoIcon
    INFORMATION = QSystemTrayIcon.MessageIcon.Information
    WARNING = QSystemTrayIcon.MessageIcon.Warning
    CRITICAL = QSystemTrayIcon.MessageIcon.Critical


class Duration(IntEnum):
    SHORT = 4
    LONG = 10


class SyncTrayMsg(NamedTuple):
    title: str
    msg: str
    duration: QSystemTrayIcon.MessageIcon
    severity: int


class TrayMessages(QObject):
    def __init__(self) -> None:
        super().__init__()

        self.messages = {
            Event.UPDATE_AVAILABLE: SyncTrayMsg(
                self.tr('An update is available'),
                self.tr('You can update the app from your digital store'),
                Severity.INFORMATION.value, Duration.SHORT
            ),
            Event.DOWNLOAD_ERROR: SyncTrayMsg(
                self.tr('The download could not be completed'),
                self.tr('There was an error while downloading your content'),
                Severity.WARNING.value, Duration.LONG
            ),
            Event.APP_RUNNING: SyncTrayMsg(
                self.tr('The app is running in the background'),
                self.tr('Click the tray icon to manage your downloads'),
                Severity.INFORMATION.value, Duration.LONG
            )
        }

    def get_msg(self, evt: Event) -> SyncTrayMsg:
        return self.messages[evt]



"""
BlackboardSync
"""

# Copyright (C) 2023, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

__all__ = [
    "__id__",
    "__title__",
    "__summary__",
    "__uri__",
    "__homepage__",
    "__author__",
    "__email__",
    "__publisher__",
    "__license__",
    "__license_spdx__",
    "__copyright__",
]

__id__ = "app.bbsync.BlackboardSync"
__title__ = "Blackboard Sync"

__summary__ = "Automatic Syncing Of Your Blackboard Content"
__uri__ = "https://github.com/sanjacob/BlackboardSync"
__homepage__ = "https://bbsync.app"

__author__ = "Jacob Sánchez"
__email__ = "jacobszpz@protonmail.com"
__publisher__ = "WeAreMagic"

__license__ = "GNU General Public License v2.0 only"
__license_spdx__ = "GPL-2.0-only"

__copyright__ = f"2024, {__author__}"



import os
import sys
sys.path.insert(0, os.path.abspath('.'))

from typing import Any
from packaging.version import Version

from jinja2 import Environment, select_autoescape
from jinja2 import FileSystemLoader

from blackboard_sync import (
    __id__,
    __title__,
    __summary__,
    __uri__,
    __homepage__,
    __author__,
    __email__,
    __publisher__,
    __license__,
    __license_spdx__,
    __copyright__
)

from releases import get_releases


def get_version_base(version: str) -> str:
    v = Version(version)
    return v.base_version


def replace_templates(template_files, context, outdir) -> int:
    folders = ["packaging/windows", "packaging/macos", "packaging/linux"]

    env = Environment(
        loader=FileSystemLoader(folders),
        autoescape=select_autoescape()
    )

    for filename in template_files:
        template = env.get_template(filename)

        outfile = f"{outdir}/{filename}"
        template.stream(**context).dump(outfile)

    return 0


def main(argv: list[str]) -> int:
    __version__ = argv[1]
    __version_base__ = get_version_base(__version__)

    template_files = [
        # Windows (MSIX)
        "AppXManifest.xml",
        "PackagingLayout.xml",

        # Windows
        "pkg_win.nsi",

        # macOS
        "pkg_macos.sh",

        # Linux
        "app.bbsync.BlackboardSync.desktop",
        "app.bbsync.BlackboardSync.metainfo.xml"
    ]

    context = {
        'package': __id__,
        'title':__title__,
        'version':__version__,
        'version_base': __version_base__,
        'author':__author__,
        'summary':__summary__,
        'homepage':__homepage__,
        'repository':__uri__,
        'publisher':__publisher__,
        'license':__license__,
        'license_spdx':__license_spdx__,
        'copyright':__copyright__,
        'releases':get_releases()
    }

    os.mkdir("replaced")
    replace_templates(template_files, context, "replaced")


if __name__ == '__main__':
    raise SystemExit(main(sys.argv))



from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBCourseContent

from .api_path import BBContentPath
from .job import DownloadJob
from . import content


class Folder:
    """Content of type `x-bb-folder`."""

    def __init__(self, _: BBCourseContent, api_path: BBContentPath,
                 job: DownloadJob) -> None:
        self.children = []
        course_id = api_path['course_id']

        for child in job.session.fetch_content_children(**api_path):
            child_path = BBContentPath(content_id=child.id,
                                       course_id=course_id)
            self.children.append(content.Content(child, child_path, job))

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        if self.children:
            path.mkdir(exist_ok=True, parents=True)

        for child in self.children:
            child.write(path, executor)

    @property
    def create_dir(self) -> bool:
        return False



# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from .api_path import BBContentPath

from .course import Course
from .content import Content
from .folder import Folder
from .document import Document
from .attachment import Attachment
from .externallink import ExternalLink
from .unhandled import Unhandled
from .body import ContentBody

__all__ = [
    'BBContentPath',
    'Course',
    'Content',
    'Folder',
    'Document',
    'Attachment',
    'ExternalLink',
    'Unhandled',
    'ContentBody'
]



"""Graphical Interface Tests"""

# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.


import pytest
import requests
from datetime import datetime, timezone

from PyQt6.QtCore import Qt
from PyQt6.QtTest import QTest
from PyQt6.QtWidgets import QApplication, QFileDialog, QDialogButtonBox

from blackboard_sync.qt import (
    SetupWizard,
    LoginWebView,
    SettingsWindow,
    SyncTrayIcon,
)

from blackboard_sync.qt.dialogs import DirDialog, Dialogs


@pytest.fixture
def settings_window():
    window = SettingsWindow()
    window.show()
    return window


@pytest.fixture
def tray_icon():
    return SyncTrayIcon()


@pytest.fixture
def make_setup_wizard(monkeypatch):
    def _make_setup_wizard(institutions: list[str]):
        wizard = SetupWizard("", institutions)

        # Don't show uni not supported dialog
        monkeypatch.setattr(Dialogs, "uni_not_supported_dialog", lambda *args: True)

        # Don't show file chooser dialog
        monkeypatch.setattr(QFileDialog, "exec", lambda *args: True)
        monkeypatch.setattr(QFileDialog, "directory", lambda *args: "/")

        # monkeypatch.setattr(wizard, "_", lambda *args: None)
        return wizard

    return _make_setup_wizard


class TestSetupWizard:
    intro = 'You are a few steps away from syncing your Blackboard Learn content straight to your device!'
    uni_selection = 'First, tell us where you study'
    location_selection = 'Where do you want files to be downloaded?'
    page_number = 4
    institutions = ['Never', 'Gonna', 'Give', 'You', 'Up']

    def test_wizard_page_number(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard([])
        qtbot.addWidget(wizard)
        assert len(wizard.pageIds()) == self.page_number

    def test_wizard_intro_label(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard([])
        qtbot.addWidget(wizard)
        assert wizard.intro_label.text() == self.intro

    def test_wizard_uni_selection_label(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard([])
        qtbot.addWidget(wizard)
        assert wizard.uni_selection_label.text() == self.uni_selection

    def test_wizard_sync_location_label(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard([])
        qtbot.addWidget(wizard)
        assert wizard.sync_location_label.text() == self.location_selection

    def test_wizard_institutions(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard(self.institutions)
        qtbot.addWidget(wizard)

        for i in range(wizard.uni_selection_box.count()):
            assert wizard.uni_selection_box.itemText(i) == self.institutions[i]

    def test_wizard_page_one(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard(self.institutions)
        # Necessary to get currentId
        wizard.show()
        qtbot.addWidget(wizard)
        assert wizard.currentId() == SetupWizard.Pages.INTRO

    def test_wizard_page_two(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard(self.institutions)
        # Necessary to get currentId
        wizard.show()
        qtbot.addWidget(wizard)
        wizard.next()
        assert wizard.currentId() == SetupWizard.Pages.INSTITUTION

    def test_wizard_valid_institution(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard(self.institutions)
        # Necessary to get currentId
        wizard.show()
        qtbot.addWidget(wizard)
        wizard.next()
        qtbot.keyClicks(wizard.uni_selection_box, self.institutions[2])
        qtbot.keyClick(wizard.uni_selection_box, Qt.Key.Key_Enter)
        assert wizard.currentId() == SetupWizard.Pages.DOWNLOAD_LOCATION

    def test_wizard_invalid_institution_enter(self, qtbot, make_setup_wizard, monkeypatch):
        wizard = make_setup_wizard(self.institutions)

        # Necessary to get currentId
        wizard.show()
        qtbot.addWidget(wizard)

        wizard.next()
        qtbot.keyClicks(wizard.uni_selection_box, 'Fake')
        qtbot.keyClick(wizard.uni_selection_box, Qt.Key.Key_Enter)
        assert wizard.currentId() == SetupWizard.Pages.INSTITUTION

    def test_wizard_invalid_institution(self, qtbot, make_setup_wizard, monkeypatch):
        wizard = make_setup_wizard(self.institutions)

        # Necessary to get currentId
        wizard.show()
        qtbot.addWidget(wizard)

        wizard.next()
        qtbot.keyClicks(wizard.uni_selection_box, 'Fake')
        wizard.next()
        assert wizard.currentId() == SetupWizard.Pages.INSTITUTION

    def test_wizard_institution_attribute(self, qtbot, make_setup_wizard):
        wizard = make_setup_wizard(self.institutions)
        # Necessary to get currentId
        wizard.show()
        qtbot.addWidget(wizard)
        wizard.next()
        qtbot.keyClicks(wizard.uni_selection_box, self.institutions[3])
        qtbot.keyClick(wizard.uni_selection_box, Qt.Key.Key_Enter)
        assert wizard.institution_index == 3


SyncPeriod = SettingsWindow.SyncPeriod

class TestSettingsWindow:
    user = 'exampleUser'
    data_source = 'exampleDataSource'

    def test_settings_window_initial_state(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)
        assert settings_window.sync_frequency == SyncPeriod.HALF_HOUR

    def test_settings_window_download_location_dialog(self, qtbot, settings_window,
                                                      tmp_path, monkeypatch):
        qtbot.addWidget(settings_window)

        monkeypatch.setattr(DirDialog, "choose", lambda *args: tmp_path)

        qtbot.mouseClick(settings_window.select_download_location, Qt.MouseButton.LeftButton)
        assert settings_window.download_location == tmp_path
        assert settings_window.download_location_hint.text() == str(tmp_path)

    def test_settings_window_download_location(self, qtbot, settings_window, tmp_path):
        qtbot.addWidget(settings_window)
        settings_window.download_location = tmp_path
        assert settings_window.download_location_hint.text() == str(tmp_path)
        assert settings_window.download_location == tmp_path

    def test_settings_window_sync_frequency(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)
        settings_window.sync_frequency = int(SyncPeriod.HALF_HOUR)
        assert settings_window.sync_frequency == SyncPeriod.HALF_HOUR

        settings_window.sync_frequency = int(SyncPeriod.ONE_HOUR)
        assert settings_window.sync_frequency == SyncPeriod.ONE_HOUR

        settings_window.sync_frequency = int(SyncPeriod.SIX_HOURS)
        assert settings_window.sync_frequency == SyncPeriod.SIX_HOURS

    def test_settings_window_sync_frequency_keys(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)

        assert settings_window.sync_frequency == SyncPeriod.HALF_HOUR
        qtbot.mouseClick(settings_window.frequency_combo, Qt.MouseButton.LeftButton)
        qtbot.keyEvent(QTest.KeyAction.Click, settings_window.frequency_combo, Qt.Key.Key_Down)
        qtbot.keyEvent(QTest.KeyAction.Click, settings_window.frequency_combo, Qt.Key.Key_Enter)
        assert settings_window.sync_frequency == SyncPeriod.ONE_HOUR
        qtbot.mouseClick(settings_window.frequency_combo, Qt.MouseButton.LeftButton)
        qtbot.keyEvent(QTest.KeyAction.Click, settings_window.frequency_combo, Qt.Key.Key_Down)
        qtbot.keyEvent(QTest.KeyAction.Click, settings_window.frequency_combo, Qt.Key.Key_Enter)
        assert settings_window.sync_frequency == SyncPeriod.SIX_HOURS

    def test_settings_window_data_source(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)
        settings_window.data_source = self.data_source
        assert settings_window.data_source == self.data_source

    def test_settings_window_username(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)
        settings_window.username = self.user
        assert self.user in settings_window.username

        settings_window.username = ''
        assert settings_window.username != ''
        assert self.user not in settings_window.username

    def test_settings_window_log_out_signal(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)

        with qtbot.waitSignal(settings_window.signals.log_out) as blocker:
            qtbot.mouseClick(settings_window.log_out_button, Qt.MouseButton.LeftButton)

        assert blocker.signal_triggered
    
    def test_settings_window_setup_wizard_signal(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)

        with qtbot.waitSignal(settings_window.signals.setup_wiz) as blocker:
            qtbot.mouseClick(settings_window.setup_button, Qt.MouseButton.LeftButton)

        assert blocker.signal_triggered

    def test_settings_window_save_signal(self, qtbot, settings_window):
        qtbot.addWidget(settings_window)
        save_btn = settings_window.button_box.button(QDialogButtonBox.StandardButton.Save)

        with qtbot.waitSignal(settings_window.signals.save) as blocker:
            qtbot.mouseClick(save_btn, Qt.MouseButton.LeftButton)

        assert blocker.signal_triggered


class TestSyncTrayIcon:
    def test_tray_icon_initial_state(self, qtbot, tray_icon):
        assert tray_icon._menu._status.text() == 'Not logged in'
        assert not tray_icon._menu.refresh.isVisible()
        assert not tray_icon._menu.preferences.isVisible()
        assert tray_icon._menu._status.isVisible()
        assert tray_icon._menu.quit.isVisible()

    def test_tray_icon_logged_in_state(self, qtbot, tray_icon):
        tray_icon.set_logged_in(True)
        assert tray_icon._menu._status.text() != 'Not logged in'
        assert tray_icon._menu.refresh.isVisible()
        assert tray_icon._menu.refresh.isEnabled()
        assert tray_icon._menu.preferences.isVisible()
        assert tray_icon._menu._status.isVisible()
        assert tray_icon._menu.quit.isVisible()

    def test_tray_icon_currently_syncing(self, qtbot, tray_icon):
        tray_icon.set_logged_in(True)
        tray_icon.set_currently_syncing(True)
        assert tray_icon._menu._status.text() == 'Downloading now...'
        assert not tray_icon._menu.refresh.isEnabled()

        tray_icon.set_currently_syncing(False)
        assert tray_icon._menu.refresh.isEnabled()

    def test_tray_icon_setupwiz_signal(self, qtbot, tray_icon):
        with qtbot.waitSignal(tray_icon.signals.reset_setup) as blocker:
            tray_icon._menu.reset_setup.trigger()

        assert blocker.signal_triggered

    def test_tray_icon_quit_signal(self, qtbot, tray_icon):
        with qtbot.waitSignal(tray_icon.signals.quit) as blocker:
            tray_icon._menu.quit.trigger()

        assert blocker.signal_triggered

    def test_tray_icon_settings_signal(self, qtbot, tray_icon):
        tray_icon.set_logged_in(True)
        with qtbot.waitSignal(tray_icon.signals.settings) as blocker:
            tray_icon._menu.preferences.trigger()

        assert blocker.signal_triggered

    def test_tray_icon_sync_signal(self, qtbot, tray_icon):
        tray_icon.set_logged_in(True)
        with qtbot.waitSignal(tray_icon.signals.force_sync) as blocker:
            tray_icon._menu.refresh.trigger()

        assert blocker.signal_triggered

    def test_tray_icon_last_synced(self, qtbot, tray_icon):
        exampleDate = datetime(year=1970, month=1, day=10, tzinfo=timezone.utc)
        tray_icon.set_last_synced(exampleDate)
        assert tray_icon._menu._last_synced == exampleDate
        assert '54 years' in tray_icon._menu._status.text()



from datetime import datetime, timezone

from blackboard.api_extended import BlackboardExtended


UNIX_EPOCH = datetime.fromtimestamp(0, tz=timezone.utc)


class DownloadJob:
    def __init__(self, session: BlackboardExtended,
                 last_downloaded: datetime | None):
        self._last_downloaded = last_downloaded or UNIX_EPOCH
        self._session = session
        self._cancelled = False

    def has_changed(self, modified: datetime | None) -> bool:
        if modified is not None:
            return (modified >= self._last_downloaded)
        return True

    @property
    def session(self) -> BlackboardExtended:
        return self._session

    @property
    def cancelled(self) -> bool:
        return self._cancelled

    def cancel(self) -> None:
        self._cancelled = True



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from enum import Enum
from pathlib import Path

from PyQt6 import uic
from PyQt6.QtCore import QObject
from PyQt6.QtGui import QIcon, QPixmap
from PyQt6.QtWidgets import QWizard


ICON_FILENAME = 'logo.png'


def load_ui(qt_obj: QObject) -> None:
    """Load a UI file for a `QObject`."""
    filename = qt_obj.__class__.__name__ + ".ui"
    filepath = Path(__file__).parent / filename
    uic.loadUi(filepath.resolve(), qt_obj)  # type: ignore


def get_asset(icon: str) -> Path:
    """Get the `Path` of a media asset."""
    root_dir = Path(__file__).parent.parent
    return (root_dir / 'assets' / icon).resolve()


def get_translations() -> str:
    qt_dir = Path(__file__).parent
    return str((qt_dir / "translations").resolve())


def get_icon(file: str) -> QIcon:
    return QIcon(str(get_asset(file)))


def logo() -> QIcon:
    return get_icon(ICON_FILENAME)


class AppIcon(Enum):
    EXIT = QIcon.ThemeIcon.ApplicationExit
    OPEN = QIcon.ThemeIcon.FolderOpen
    SYNC = QIcon.ThemeIcon.ViewRefresh
    HELP = QIcon.ThemeIcon.HelpAbout
    HOME = QIcon.ThemeIcon.GoHome
    BACK = QIcon.ThemeIcon.EditUndo


def get_theme_icon(icon: AppIcon) -> QIcon:
    return QIcon.fromTheme(icon.value)


def get_wizard_pixmap(which: QWizard.WizardPixmap) -> QPixmap:
    return QPixmap(str(get_asset(f"{which}.png")))



#!/usr/bin/env python3

"""BlackboardSync Controller."""

# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from requests.cookies import RequestsCookieJar
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_version

from .sync import BlackboardSync
from .__about__ import __id__, __title__, __uri__
from .institutions import get_names, autodetect

from .updates import check_for_updates
from .qt.manager import UIManager


class SyncController:
    """Connects an instance of BlackboardSync with the UI module."""

    def __init__(self):
        super().__init__()
        self.model = BlackboardSync()
        self.ui = UIManager(__id__, __title__, __uri__,
                            get_names(), autodetect())

        first_time = self.model.university is None

        if not first_time:
            self.open_login()

        self.init_signals()
        self.ui.start(first_time)

    def init_signals(self):
        self.ui.signals.open_settings.connect(self.open_settings)
        self.ui.signals.open_tray.connect(self.open_tray)
        self.ui.signals.open_downloads.connect(self.open_downloads)
        self.ui.signals.open_menu.connect(self.open_menu)
        self.ui.signals.setup.connect(self.setup)
        self.ui.signals.config.connect(self.config)
        self.ui.signals.redownload.connect(self.redownload)
        self.ui.signals.force_sync.connect(self.force_sync)
        self.ui.signals.log_in.connect(self.log_in)
        self.ui.signals.log_out.connect(self.log_out)
        self.ui.signals.quit.connect(self.quit)

    def open_login(self) -> None:
        start_url = str(self.model.university.login.start_url)
        target_url = str(self.model.university.login.target_url)

        self.ui.open_login(start_url, target_url)

    def force_sync(self) -> None:
        self.model.force_sync()

    def open_settings(self) -> None:
        __version__ = None
        package = __package__.replace('_', '')

        try:
            __version__ = get_version(package)
        except PackageNotFoundError:
            pass

        self.ui.open_settings(self.model.download_location,
                              self.model.username,
                              self.model.sync_interval,
                              __version__)

    def open_menu(self) -> None:
        self.ui.open_menu(self.model.last_sync_time,
                          self.model.is_logged_in,
                          self.model.is_syncing)

        if self.model.has_error:
            self.ui.notify_sync_error()

    def open_tray(self, clicked) -> None:
        if clicked:
            first_time = self.model.university is None
            is_logged_in = self.model.is_logged_in

            self.ui.open_tray(first_time, is_logged_in)

    def open_downloads(self) -> None:
        self.ui.open_file(self.model.download_location)

    def setup(self, institution_index: int,
              download_location: str, min_year: int) -> None:
        self.model.setup(institution_index, download_location, min_year)
        self.open_login()

    def config(self, download_location: str, sync_frequency: int) -> None:
        if self.model.download_location != download_location:
            self.model.download_location = download_location
            self.ui.ask_redownload()

        self.model.sync_interval = sync_frequency

    def redownload(self) -> None:
        self.model.redownload()

    def log_in(self, cookies: RequestsCookieJar) -> None:
        if self.model.auth(cookies):
            self.ui.log_in()
            self.ui.notify_running()
            self.check_for_updates()
        else:
            self.ui.notify_login_error()

    def log_out(self) -> None:
        if self.model.is_active:
            self.model.stop_sync()

        self.model.log_out()

    def quit(self) -> None:
        if self.model.is_active:
            self.model.stop_sync()

    def check_for_updates(self) -> None:
        if check_for_updates():
            self.ui.notify_update()


if __name__ == '__main__':
    controller = SyncController()



import logging
from pathlib import Path
from json import JSONDecodeError
from pydantic import ValidationError
from requests import RequestException
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import (
    BBCourseContent,
    BBResourceType,
    BBContentHandler
)
from blackboard.exceptions import BBBadRequestError, BBForbiddenError

from . import folder, document, externallink, body, unhandled

from .api_path import BBContentPath
from .job import DownloadJob

logger = logging.getLogger(__name__)


class Content:
    """Content factory for all types."""

    def __init__(self, content: BBCourseContent, api_path: BBContentPath,
                 job: DownloadJob) -> None:

        logger.info(f"{content.title}[{content.contentHandler}]")

        self.body = None
        self.handler = None

        self.ignore = not Content.should_download(content, job)

        if self.ignore:
            return

        Handler = Content.get_handler(content.contentHandler)
        self.title = content.title_path_safe.replace('.', '_')

        try:
            self.handler = Handler(content, api_path, job)
        except (ValidationError, JSONDecodeError,
                BBBadRequestError, BBForbiddenError):
            logger.exception(f"Error fetching {content.title}")

        try:
            if content.body:
                self.body = body.ContentBody(content, None, job)
        except (ValidationError, JSONDecodeError,
                BBBadRequestError, BBForbiddenError, RequestException):
            logger.warning(f"Error fetching body of {content.title}")

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        if self.ignore:
            return

        # Build nested path with content title
        path = path / self.title

        if self.handler is not None:
            if self.handler.create_dir:
                path.mkdir(exist_ok=True, parents=True)

            self.handler.write(path, executor)

        if self.body is not None:
            path.mkdir(exist_ok=True, parents=True)
            self.body.write(path, executor)

    @staticmethod
    def should_download(content: BBCourseContent, job: DownloadJob) -> bool:
        or_guards = [
            job.has_changed(content.modified),
            content.hasChildren,
        ]

        return any(or_guards) and bool(content.availability)

    @staticmethod
    def get_handler(content_handler: BBContentHandler | None):
        match content_handler:
            case BBResourceType.Folder | BBResourceType.Lesson:
                return folder.Folder
            case BBResourceType.File | BBResourceType.Document:
                return document.Document
            case BBResourceType.Assignment:
                return document.Document
            case BBResourceType.ExternalLink:
                return externallink.ExternalLink
            case _:
                return unhandled.Unhandled



"""Loads and provides data about supported universities."""

# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import json
import logging
from pathlib import Path

from pydantic import HttpUrl, BaseModel

from .ip import find_my_ip, find_ip_entity


__all__ = [
    'InstitutionLogin',
    'InstitutionNetwork',
    'InstitutionContent',
    'Institution',
    'get_names',
    'get_by_index',
    'autodetect'
]


logger = logging.getLogger(__name__)

UNIVERSITY_DB = 'universities.json'


class InstitutionLogin(BaseModel):
    start_url: HttpUrl
    target_url: HttpUrl


class InstitutionNetwork(BaseModel):
    name: list[str] = []


class InstitutionContent(BaseModel):
    data_sources: list[str] = []


class Institution(BaseModel):
    """A university/college with a Blackboard Learn instance."""

    name: str
    """Official university name."""

    short_name: str | None = None
    """Abbreviation used to facilitate search."""

    country: str | None = None
    """Country of university."""

    api_url: HttpUrl
    """URL of the Blackboard Learn REST API instance."""

    login: InstitutionLogin
    """Metadata to make login possible."""

    network: InstitutionNetwork | None = None
    """Metadata to enable automatic detection."""

    content: InstitutionContent | None = None
    """Metadata to improve content filtering."""


def load() -> list[Institution]:
    db = []
    db_file = Path(__file__).parent / UNIVERSITY_DB

    with db_file.open(encoding='utf-8') as f:
        db = json.load(f)

    return [Institution(**uni) for uni in db]


_institutions = load()


def get_by_index(i: int) -> Institution:
    """Get the institution with the given id."""
    return _institutions[i]


def get_names() -> list[str]:
    """Get university names for search."""
    names = []

    for uni in _institutions:
        tag = f"{uni.name} ({uni.short_name})" if uni.short_name else uni.name
        names.append(tag)
    return names


def autodetect() -> int | None:
    """Detect university based on ISP information."""

    my_ip = find_my_ip()

    if my_ip is None:
        return None

    entity = find_ip_entity(my_ip)

    if entity is None:
        return None

    name = entity['name']
    description = entity['description']

    for i, uni in enumerate(_institutions):
        # First heuristic, network description
        # Works only with full university name
        if description and description[0] == uni.name:
            return i

        # Second heuristic, network name
        # More specific but must be specified manually
        if uni.network and any(name == x for x in uni.network.name):
            return i
    return None



"""
Blackboard Sync

Download your Blackboard Learn content automatically.
"""

# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import time
import logging
import threading
from pathlib import Path
from requests import RequestException
from datetime import datetime, timezone, timedelta

from requests.cookies import RequestsCookieJar

from blackboard.api_extended import BlackboardExtended
from blackboard.exceptions import BBUnauthorizedError, BBForbiddenError

from .config import SyncConfig
from .download import BlackboardDownload
from .institutions import Institution, get_by_index

logger = logging.getLogger(__name__)


class BlackboardSync:
    """Represents an instance of the BlackboardSync application."""

    _log_directory = "log"

    # Seconds between each check of time elapsed since last sync
    _check_sleep_time = 10

    def __init__(self) -> None:
        """Create an instance of the program."""

        # Download job
        self._download: BlackboardDownload | None = None

        # Time between each sync in seconds
        self._sync_interval = 60 * 30

        # Time of next programmed sync
        self._next_sync = None
        # User session active
        self._is_logged_in = False

        # Flag to force sync
        self._force_sync = False
        # Flag to know if syncing is in progress
        self._is_syncing = False
        # Flag to know if syncing is on
        self._is_active = False
        # Flag to know if download thread has errors
        self._has_error = False

        logger.debug("Initialising BlackboardSync")

        self.university: Institution | None = None
        self.sess: BlackboardExtended | None = None

        # Attempt to load existing configuration
        self._config = SyncConfig()
        logger.info("Loading preexisting configuration")

        if self._config.university_index is not None:
            self.university = get_by_index(self._config.university_index)

        if self._config.last_sync_time is not None:
            self.schedule_next_sync(self._config.last_sync_time)

        if self.download_location is not None:
            self._add_logger_file_handler()

    def setup(self, university_index: int, download_location: Path,
              min_year: int | None = None) -> None:
        """Setup the university information."""
        self.university_index = university_index
        self.download_location = download_location

        # If min_year has decreased, redownload
        if (self._config.min_year or 0) > (min_year or 0):
            self.redownload()

        self._config.min_year = min_year

    def auth(self, cookies: RequestsCookieJar) -> bool:
        """Create a new Blackboard session with the given cookies."""
        if self.university is None:
            return False

        api_url = str(self.university.api_url)

        try:
            u_sess = BlackboardExtended(api_url, cookies=cookies)
            # should trigger exception if not authenticated
            u_sess.fetch_users(user_id='me')
        except (BBUnauthorizedError, BBForbiddenError):
            logger.warning("Credentials are incorrect")
        except RequestException:
            logger.warning("Error while making auth request")
        else:
            logger.info("Logged in successfully")
            self.sess = u_sess
            self._is_logged_in = True
            self.start_sync()

        return self._is_logged_in

    def log_out(self) -> None:
        """Stop syncing and forget user session."""
        self.stop_sync()
        self.sess = None
        self._is_logged_in = False

    def download(self) -> datetime | None:
        user_session = self.sess

        if user_session is None or self.university is None:
            return None

        if self.download_location is None:
            return None

        self._download = BlackboardDownload(
            user_session,
            self.download_location,
            self.last_sync_time,
            self.min_year
        )

        if not self._is_active:
            return None

        try:
            start_time = self._download.download()
        except BBUnauthorizedError:
            logger.exception("User session expired")
            self.log_out()
        except Exception:
            logger.exception("Download error")
            self._has_error = True

            # manually postpone next sync job
            self.schedule_next_sync(datetime.now(timezone.utc))
        else:
            return start_time
        return None

    def _sync_task(self) -> None:
        """Constantly check if data is outdated and if so start download.

        Method run by Sync thread.
        """
        while self._is_active:
            if self.outdated or self._force_sync:
                logger.info("Syncing now")
                self._is_syncing = True

                start_time = self.download()

                if start_time is not None:
                    self.last_sync_time = start_time

                # Reset sync flags
                self._force_sync = False
                self._is_syncing = False

            if self._is_active:
                time.sleep(self._check_sleep_time)

    def start_sync(self) -> bool:
        """Starts Sync thread or returns False if not possible."""
        if self._has_error:
            return False

        logger.info("Starting sync thread")
        self._is_active = True
        self.sync_thread = threading.Thread(target=self._sync_task)
        self.sync_thread.start()
        return True

    def stop_sync(self) -> None:
        """Stop Sync thread."""
        logger.info("Stopping sync thread")
        self._is_active = False

        if self._download is not None:
            self._download.cancel()

    def _add_logger_file_handler(self) -> None:
        filename = f"sync_log_{datetime.now():%Y-%m-%d}.log"

        if self.download_location is None:
            return

        log_dir = self.download_location / self._log_directory
        log_dir.mkdir(exist_ok=True, parents=True)

        log_path = log_dir / filename

        logger = logging.getLogger(__name__)

        file_handler = logging.FileHandler(log_path)
        file_handler.setLevel(logging.WARNING)

        logger.addHandler(file_handler)

    def force_sync(self) -> None:
        """Force Sync thread to start download job ASAP."""
        logger.debug("Forced syncing")
        self._force_sync = True

    def redownload(self) -> None:
        self.last_sync_time = None

    @property
    def username(self) -> str | None:
        return self.sess.user_id if self.sess is not None else None

    @property
    def last_sync_time(self) -> datetime | None:
        """Datetime right before last download job started."""
        return self._config.last_sync_time

    @last_sync_time.setter
    def last_sync_time(self, last_time: datetime | None) -> None:
        """Updates the last sync time recorded."""
        self._config.last_sync_time = last_time
        self.schedule_next_sync(last_time)

    def schedule_next_sync(self, start_time: datetime | None) -> None:
        if start_time is not None:
            delay = timedelta(seconds=self._sync_interval)
            self._next_sync = start_time + delay

    @property
    def next_sync(self) -> datetime | None:
        """Time when last sync will be outdated."""
        return self._next_sync

    @property
    def outdated(self) -> bool:
        """Return true if last download job is outdated."""
        if self.next_sync is None:
            return True
        return datetime.now(timezone.utc) >= self.next_sync

    @property
    def min_year(self) -> int | None:
        return self._config.min_year

    @property
    def university_index(self) -> int | None:
        return self._config.university_index

    @university_index.setter
    def university_index(self, uni_index: int) -> None:
        self._config.university_index = uni_index
        self.university = get_by_index(uni_index)

    @property
    def download_location(self) -> Path | None:
        """Location to where all the content will be downloaded."""
        return self._config.download_location

    @download_location.setter
    def download_location(self, value: Path) -> None:
        if value != self.download_location:
            self._config.download_location = value
            self._add_logger_file_handler()

    @property
    def sync_interval(self) -> int:
        """Time to wait between download jobs."""
        return self._sync_interval

    @sync_interval.setter
    def sync_interval(self, p: int) -> None:
        self._sync_interval = p

    @property
    def is_active(self) -> bool:
        """Indicate the state of the sync thread."""
        return self._is_active

    @property
    def is_logged_in(self) -> bool:
        """Indicate if a user session is currently active."""
        return self._is_logged_in

    @property
    def is_syncing(self) -> bool:
        """Flag raised everytime a download job is running."""
        return self._is_syncing

    @property
    def has_error(self) -> bool:
        """Flag indicates an error resulting in no downloads."""
        if self._has_error:
            self._has_error = False
            return True
        return False



from pathlib import Path
from unittest import mock

from hypothesis import given
from hypothesis import assume

from blackboard.blackboard import (
    BBCourse,
    BBCourseContent,
    BBContentHandler,
    BBResourceType,
    BBAttachment,
    BBAvailability
)

from blackboard_sync.content import (
    BBContentPath,
    Course,
    Content,
    Folder,
    Document,
    Unhandled
)

from blackboard_sync.content.base import FStream


def assert_written(path, content):
    assert path.is_file()
    with path.open('r') as f:
        assert content == f.read()


def make_available(model):
    return model.model_copy(update={
        'availability': BBAvailability(available='Yes')
    })


def get_module(class_name):
    return 'blackboard_sync.content.content.' + class_name


def test_fstream_write(tmpdir):
    new_file_path = Path(tmpdir, 'my_new_file.txt')
    contents = 'whatever the file says\nincluding newlines'

    # Write a text file to disk
    assert not new_file_path.exists()

    content = FStream()
    content.write_base(new_file_path, mock.Mock(submit=lambda x: x()),
                       contents)

    assert new_file_path.is_file()

    with new_file_path.open('r') as f:
        assert contents == f.read()


@given(...)
def test_course_api_call(course: BBCourse):
    course = make_available(course)

    job = mock.Mock()
    job.session.fetch_contents.return_value = []

    Course(course, job)

    job.session.fetch_contents.assert_called_once_with(
        course_id=course.id
    )


@given(...)
def test_folder_api_call(api_path: BBContentPath):
    job = mock.Mock()
    job.session.fetch_content_children.return_value = []

    Folder(None, api_path, job)
    job.session.fetch_content_children.assert_called_once_with(**api_path)


@given(...)
def test_file_api_call(api_path: BBContentPath):
    job = mock.Mock()
    job.session.fetch_file_attachments.return_value = []

    Document(None, api_path, job)

    job.session.fetch_file_attachments.assert_called_once_with(
        **api_path
    )


@given(...)
def test_children_course(course: BBCourse, contents: list[BBCourseContent]):
    job = mock.Mock()
    job.session.fetch_contents.return_value = contents

    course = make_available(course)

    calls = []
    for content in contents:
        api_path = BBContentPath(course_id=course.id, content_id=content.id)
        calls.append(mock.call(content, api_path, job))

    with mock.patch('blackboard_sync.content.course.Content') as p:
        Course(course, job)
        p.assert_has_calls(calls)


@given(...)
def test_children_folder(api_path: BBContentPath,
                         children: list[BBCourseContent]):
    job = mock.Mock()
    job.session.fetch_content_children.return_value = children

    calls = []
    for child in children:
        api_path = BBContentPath(course_id=api_path.get('course_id'),
                                 content_id=child.id)
        calls.append(mock.call(child, api_path, job))

    with mock.patch('blackboard_sync.content.folder.content.Content') as p:
        Folder(None, api_path, job)
        p.assert_has_calls(calls)


@given(...)
def test_children_file(api_path: BBContentPath,
                       attachments: list[BBAttachment]):
    job = mock.Mock()
    job.session.fetch_file_attachments.return_value = attachments

    calls = []

    # Setup calls
    for att in attachments:
        if att.mimeType is not None:
            assume(not att.mimeType.startswith("video/"))
            assume(not att.mimeType == '*')

        calls.append(mock.call(att, api_path, job))

    with mock.patch('blackboard_sync.content.document.Attachment') as p:
        Document(None, api_path, job)
        p.assert_has_calls(calls)


@given(...)
def test_content_folder(api_path: BBContentPath):
    job = mock.Mock()
    job.has_changed.return_value = True

    content = mock.MagicMock()
    content.contentHandler = BBContentHandler(id=BBResourceType.Folder)
    content.availability = True
    content.id = api_path['content_id']
    content.body = False

    with mock.patch(get_module('folder.Folder')) as p:
        Content(content, api_path, job)
        p.assert_called_once_with(content, api_path, job)


@given(...)
def test_content_file(api_path: BBContentPath):
    job = mock.Mock()
    job.has_changed.return_value = True

    content = mock.MagicMock()
    content.contentHandler = BBContentHandler(id=BBResourceType.Document)
    content.availability = True
    content.id = api_path['content_id']
    content.body = False

    with mock.patch(get_module('document.Document')) as p:
        Content(content, api_path, job)
        p.assert_called_once_with(content, api_path, job)


@given(...)
def test_content_link(api_path: BBContentPath):
    job = mock.Mock()
    job.has_changed.return_value = True

    content = mock.MagicMock()
    content.contentHandler = BBContentHandler(id=BBResourceType.ExternalLink)
    content.availability = True
    content.id = api_path['content_id']
    content.body = False

    with mock.patch(get_module('externallink.ExternalLink')) as p:
        Content(content, api_path, job)
        p.assert_called_once_with(content, api_path, job)


@given(...)
def test_content_unhandled(api_path: BBContentPath):
    job = mock.Mock()
    job.has_changed.return_value = True

    content = mock.MagicMock()
    content.contentHandler = BBContentHandler(id=BBResourceType.Blankpage)
    content.availability = True
    content.id = api_path['content_id']
    content.body = False

    with mock.patch(get_module('unhandled.Unhandled')) as p:
        Content(content, api_path, job)
        p.assert_called_once_with(content, api_path, job)


def test_unhandled():
    executor = mock.Mock()

    content = Unhandled(mock.Mock(title="Content title"), None, mock.Mock())
    content.write(None, executor)
    executor.submit.assert_not_called()



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import sys
from pathlib import Path
from datetime import datetime
from requests.cookies import RequestsCookieJar
from importlib.metadata import version, PackageNotFoundError

from PyQt6.QtCore import QLocale, QTranslator, QLibraryInfo
from PyQt6.QtCore import Qt, pyqtSlot, pyqtSignal, QObject
from PyQt6.QtWidgets import QApplication, QSystemTrayIcon, QWidget

from . import SetupWizard, LoginWebView, SettingsWindow, SyncTrayIcon
from .notification import Event
from .dialogs import Dialogs
from .utils import add_to_startup, open_in_file_browser, windows_safe_path
from .assets import get_translations


class UIManager(QObject):
    class Signals(QObject):
        open_settings = pyqtSignal()
        open_tray = pyqtSignal(bool)
        open_menu = pyqtSignal()
        open_downloads = pyqtSignal()
        setup = pyqtSignal(int, Path, int)
        config = pyqtSignal(Path, int)
        force_sync = pyqtSignal()
        log_out = pyqtSignal()
        log_in = pyqtSignal(RequestsCookieJar)
        redownload = pyqtSignal()
        quit = pyqtSignal()

    def __init__(self, id: str, title: str, uri: str,
                 universities: list[str], autodetected: int | None) -> None:
        super().__init__()
        self.id = id
        self.title = title
        self.uri = uri
        self.help_uri = f"{uri}/issues"

        self.signals = self.Signals()

        self.app = QApplication(sys.argv)
        self.app.setApplicationName(title)
        self.app.setQuitOnLastWindowClosed(False)
        self.load_translator()

        try:
            __version__ = version(Path(__file__).parent.parent.stem)
        except PackageNotFoundError:
            pass
        else:
            self.app.setApplicationVersion(__version__)

        self._init_ui(universities, autodetected)

    def _init_ui(self, universities: list[str],
                 autodetected: int | None) -> None:
        self.setup_window = SetupWizard(self.help_uri,
                                        universities, autodetected)
        self.login_window = LoginWebView(self.help_uri)
        self.config_window = SettingsWindow()
        self.tray = SyncTrayIcon()
        self.dialogs = Dialogs()

        self.login_window.signals.login_complete.connect(self.slot_log_in)
        self.setup_window.accepted.connect(self.slot_setup)

        self.config_window.signals.log_out.connect(self.signals.log_out)
        self.config_window.signals.save.connect(self.slot_config)
        self.config_window.signals.log_out.connect(self.slot_log_out)
        self.config_window.signals.setup_wiz.connect(self.slot_open_setup)

        self.tray.signals.force_sync.connect(self.signals.force_sync)
        self.tray.signals.show_menu.connect(self.signals.open_menu)
        self.tray.signals.settings.connect(self.signals.open_settings)
        self.tray.signals.open_dir.connect(self.signals.open_downloads)
        self.tray.signals.quit.connect(self.signals.quit)
        self.tray.signals.reset_setup.connect(self.slot_open_setup)
        self.tray.activated.connect(self.slot_open_tray)
        self.tray.signals.quit.connect(self.slot_quit)

    def load_translator(self) -> None:
        tr_path = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
        self.locale = QLocale.system().name()

        translations = [
            ('qtbase_{}', tr_path),
            ('qtwebengine_{}', tr_path),
            ('{}', get_translations())
        ]

        self.translators = []

        for filename, path in translations:
            translator = QTranslator()

            if translator.load(filename.format(self.locale), path):
                self.app.installTranslator(translator)
                self.translators.append(translator)

    def start(self, first_time: bool) -> None:
        if first_time:
            add_to_startup(self.id)
            self.show(self.setup_window)
        else:
            self.show(self.login_window)

        self.app.exec()

    def show(self, widget: QWidget) -> None:
        widget.setWindowState(Qt.WindowState.WindowNoState)
        widget.show()
        widget.setFocus()

    def hide(self, widget: QWidget) -> None:
        widget.setVisible(False)

    @pyqtSlot(QSystemTrayIcon.ActivationReason)
    def slot_open_tray(self,
                       reason: QSystemTrayIcon.ActivationReason) -> None:
        clicked = reason == QSystemTrayIcon.ActivationReason.Trigger
        self.signals.open_tray.emit(clicked)

    @pyqtSlot()
    def slot_log_out(self) -> None:
        self.tray.set_logged_in(False)
        self.login_window.restore()
        self.show(self.login_window)
        self.hide(self.config_window)

    @pyqtSlot()
    def slot_open_setup(self) -> None:
        self.signals.log_out.emit()
        self.hide(self.login_window)
        self.hide(self.config_window)
        self.show(self.setup_window)

    @pyqtSlot()
    def slot_log_in(self) -> None:
        self.signals.log_in.emit(self.login_window.cookies)

    def log_in(self) -> None:
        self.hide(self.login_window)
        self.tray.set_logged_in(True)

    @pyqtSlot()
    def slot_setup(self) -> None:
        self.hide(self.setup_window)

        sync_dir = windows_safe_path(self.setup_window.download_location)
        self.signals.setup.emit(self.setup_window.institution_index,
                                sync_dir, self.setup_window.min_year or 0)

    @pyqtSlot()
    def slot_config(self) -> None:
        self.hide(self.config_window)

        sync_dir = windows_safe_path(self.config_window.download_location)
        self.signals.config.emit(sync_dir, self.config_window.sync_frequency)

    @pyqtSlot()
    def slot_quit(self) -> None:
        self.login_window.shutdown()
        self.app.quit()

    def open_settings(self, download_location: Path, username: str,
                      sync_interval: int, version: str | None) -> None:
        self.config_window.download_location = download_location
        self.config_window.username = username
        self.config_window.sync_frequency = sync_interval
        self.config_window.version = version
        self.show(self.config_window)

    def open_menu(self, last_sync: datetime,
                  is_logged: bool, is_syncing: bool) -> None:
        self.tray.set_last_synced(last_sync)
        self.tray.set_logged_in(is_logged)
        self.tray.set_currently_syncing(is_syncing)

    def open_tray(self, first_time: bool, is_logged: bool) -> None:
        if first_time:
            self.show(self.setup_window)
        elif not is_logged:
            self.show(self.login_window)

    def open_file(self, file: Path) -> None:
        open_in_file_browser(file)

    def open_login(self, start_url: str, target_url: str) -> None:
        self.login_window.load(start_url, target_url)
        self.show(self.login_window)

    def ask_redownload(self) -> None:
        if self.dialogs.redownload_dialog():
            self.signals.redownload.emit()

    def notify_running(self) -> None:
        self.tray.notify(Event.APP_RUNNING)

    def notify_login_error(self) -> None:
        self.dialogs.login_error_dialog(self.help_uri)
        self.login_window.restore()
        self.show(self.login_window)

    def notify_sync_error(self) -> None:
        self.tray.notify(Event.DOWNLOAD_ERROR)

    def notify_update(self) -> None:
        self.tray.notify(Event.UPDATE_AVAILABLE)



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import webbrowser
from pathlib import Path

from PyQt6.QtCore import QObject
from PyQt6.QtWidgets import QMessageBox, QFileDialog

from .assets import logo


class DirDialog(QFileDialog):
    """Open the file dialog in directory mode."""
    def init(self) -> None:
        super().__init__()

    def choose(self) -> Path | None:
        self.setFileMode(QFileDialog.FileMode.Directory)
        if super().exec():
            return Path(self.directory().path())
        return None


class Dialogs(QObject):
    def __init__(self) -> None:
        super().__init__()

    def redownload_dialog(self) -> bool:
        q = QMessageBox()
        q.setText(self.tr("Do you wish to redownload all files?"))
        q.setInformativeText(self.tr(
            "Answer no if you intend to move all past downloads manually"
            " (Recommended)."
        ))
        q.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        q.setDefaultButton(QMessageBox.StandardButton.No)
        q.setIcon(QMessageBox.Icon.Question)
        q.setWindowIcon(logo())
        return q.exec() == QMessageBox.StandardButton.Yes

    def uni_not_supported_dialog(self, url: str) -> None:
        q = QMessageBox()
        q.setText(self.tr(
            "Unfortunately, your university is not yet supported"
        ))
        q.setInformativeText(self.tr(
            "You can help us provide support for it by visiting our website, "
            "which you can access by pressing the help button."
        ))
        q.setStandardButtons(
            QMessageBox.StandardButton.Help | QMessageBox.StandardButton.Ok
        )
        q.setIcon(QMessageBox.Icon.Warning)
        q.setWindowIcon(logo())

        if q.exec() == QMessageBox.StandardButton.Help:
            webbrowser.open(url)

    def login_error_dialog(self, url: str) -> None:
        q = QMessageBox()
        q.setText(self.tr(
            "There was an issue logging you in"
        ))
        q.setInformativeText(self.tr(
            "Please try again later, and if the error persists"
            " contact our support by pressing the button below."
        ))
        q.setStandardButtons(
            QMessageBox.StandardButton.Help | QMessageBox.StandardButton.Ok
        )
        q.setIcon(QMessageBox.Icon.Warning)
        q.setWindowIcon(logo())

        if q.exec() == QMessageBox.StandardButton.Help:
            webbrowser.open(url)



from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBCourseContent
from blackboard.filters import BBAttachmentFilter
from bwfilters import BWFilter

from .attachment import Attachment
from .api_path import BBContentPath
from .job import DownloadJob


class Document:
    """Represents a file with attachments in the Blackboard API"""
    def __init__(self, content: BBCourseContent, api_path: BBContentPath,
                 job: DownloadJob):
        attachments = job.session.fetch_file_attachments(**api_path)
        assert isinstance(attachments, list)

        att_filter = BBAttachmentFilter(mime_types=BWFilter(['video/*']))
        filtered_attachments = list(att_filter.filter(attachments))

        self.attachments = []

        for i, attachment in enumerate(filtered_attachments):
            self.attachments.append(
                Attachment(attachment, api_path, job)
            )

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        # If only attachment, just use parent
        if len(self.attachments) > 1:
            path.mkdir(exist_ok=True, parents=True)
        else:
            path = path.parent

        for attachment in self.attachments:
            attachment.write(path, executor)

    @property
    def create_dir(self) -> bool:
        return False



import uuid
import mimetypes

from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBAttachment

from .base import BStream
from .api_path import BBContentPath
from .job import DownloadJob


class Attachment(BStream):
    """File attached to a content."""

    def __init__(self, attachment: BBAttachment, api_path: BBContentPath,
                 job: DownloadJob):
        filename = attachment.fileName or str(uuid.uuid1())
        name_ext = '.' + filename.split('.')[-1]

        # Guess extension based on content
        mime = attachment.mimeType or 'text/plain'
        possible_ext = mimetypes.guess_all_extensions(mime, strict=False)

        if name_ext in possible_ext:
            self.filename = filename
        else:
            real_ext = possible_ext[0] if possible_ext else '.txt'
            self.filename = filename + real_ext

        self.stream = job.session.download(attachment_id=attachment.id,
                                           **api_path)

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        super().write_base(path / self.filename, executor, self.stream)



"""
BlackboardSync Qt GUI.

~~~~~~~~~~~~~~~~~~~~~

A graphical interface using Qt for BlackboardSync.
Full documentation available at <https://blackboardsync.readthedocs.io>.

:copyright: (c) 2024, Jacob Sánchez Pérez.
:license: GPL v2.0, see LICENSE for more details.
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from .SyncTrayIcon import SyncTrayIcon
from .SetupWizard import SetupWizard
from .LoginWebView import LoginWebView
from .SettingsWindow import SettingsWindow


__all__ = [
    'SetupWizard',
    'LoginWebView',
    'SettingsWindow',
    'SyncTrayIcon'
]



import logging
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBCourseContent

from .job import DownloadJob

logger = logging.getLogger(__name__)


class Unhandled:
    """Content which we cannot handle yet, or at all."""

    def __init__(self, content: BBCourseContent, _: None,
                 job: DownloadJob) -> None:
        logger.info(f"{content.title} not supported")

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        pass

    @property
    def create_dir(self) -> bool:
        return False



import logging
from pathlib import Path
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor

from blackboard.blackboard import BBCourse

from .api_path import BBContentPath
from .job import DownloadJob
from .content import Content

logger = logging.getLogger(__name__)


class Course:
    def __init__(self, course: BBCourse, job: DownloadJob):
        logger.info(f"<{course.code}> - <{course.title}>")

        self.ignore = not course.availability

        if self.ignore:
            return

        self.year = self.get_year(course.created)
        self.title = course.title or 'Untitled Course'

        contents = job.session.fetch_contents(course_id=course.id)
        self.children = []

        for content in contents:
            api_path = BBContentPath(course_id=course.id,
                                     content_id=content.id)
            self.children.append(Content(content, api_path, job))

    def write(self, path: Path, executor: ThreadPoolExecutor) -> None:
        if self.ignore:
            return

        path = path / self.year / self.title

        for child in self.children:
            child.write(path, executor)

    @staticmethod
    def get_year(created: datetime | None) -> str:
        return str(created.year) if created is not None else 'No Date'



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from datetime import datetime

from PyQt6.QtGui import QAction
from PyQt6.QtCore import pyqtSignal, QObject
from PyQt6.QtWidgets import QMenu, QSystemTrayIcon

from .assets import logo, get_theme_icon, AppIcon
from .utils import time_ago
from .notification import Event, TrayMessages


class SyncTrayMenu(QMenu):

    def __init__(self, logged_in: bool = False,
                 last_synced: datetime | None = None):
        super().__init__()
        self._last_synced: datetime | None = None

        self._init_ui()
        self.set_last_synced(last_synced)
        self.set_logged_in(logged_in)

    def _init_ui(self) -> None:
        sync_icon = get_theme_icon(AppIcon.SYNC)
        close_icon = get_theme_icon(AppIcon.EXIT)
        open_dir_icon = get_theme_icon(AppIcon.OPEN)

        self.refresh = QAction(self.tr("Sync now"))
        self.refresh.setIcon(sync_icon)
        self.addAction(self.refresh)

        self.open_dir = QAction(self.tr("Open downloads"))
        self.open_dir.setIcon(open_dir_icon)
        self.addAction(self.open_dir)

        self.preferences = QAction(self.tr("Preferences"))
        self.addAction(self.preferences)

        self.addSeparator()

        self._status = QAction(self.tr("You haven't logged in"))
        self._status.setEnabled(False)
        self.addAction(self._status)

        self.reset_setup = QAction(self.tr("Setup"))
        self.addAction(self.reset_setup)

        self.quit = QAction(self.tr("Quit"))
        self.quit.setIcon(close_icon)
        self.addAction(self.quit)

    def set_logged_in(self, logged_in: bool) -> None:
        self.refresh.setVisible(logged_in)
        self.preferences.setVisible(logged_in)
        self.reset_setup.setVisible(not logged_in)
        self.open_dir.setVisible(logged_in)

        if logged_in:
            self.set_last_synced(self._last_synced)
        else:
            self._status.setText(self.tr("Not logged in"))

    def set_last_synced(self, last_synced: datetime | None) -> None:
        self._last_synced = last_synced
        human_ago = time_ago(last_synced) if last_synced else "Never"
        self._status.setText(self.tr("Last synced: ") + human_ago)

    def set_currently_syncing(self, syncing: bool) -> None:
        self.refresh.setEnabled(not syncing)

        if syncing:
            self._status.setText(self.tr("Downloading now..."))


class SyncTrayIcon(QSystemTrayIcon):
    """Control the application from the system tray."""

    class Signals(QObject):
        force_sync = pyqtSignal()
        settings = pyqtSignal()
        reset_setup = pyqtSignal()
        quit = pyqtSignal()
        open_dir = pyqtSignal()
        show_menu = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()
        self.signals = self.Signals()

        self._init_ui()

    def _init_ui(self) -> None:
        # Create the tray
        self.setIcon(logo())
        self.setVisible(True)
        self.setToolTip("")

        # Create the menu
        self._menu = SyncTrayMenu()
        self._messages = TrayMessages()

        # Signals
        self._menu.refresh.triggered.connect(self.signals.force_sync)
        self._menu.preferences.triggered.connect(self.signals.settings)
        self._menu.reset_setup.triggered.connect(self.signals.reset_setup)
        self._menu.quit.triggered.connect(self.signals.quit)
        self._menu.open_dir.triggered.connect(self.signals.open_dir)
        self._menu.aboutToShow.connect(self.signals.show_menu)

        # Add the menu to the tray
        self.setContextMenu(self._menu)

    def set_logged_in(self, value: bool) -> None:
        self._menu.set_logged_in(value)

    def set_last_synced(self, value: datetime | None) -> None:
        self._menu.set_last_synced(value)

    def set_currently_syncing(self, syncing: bool) -> None:
        self._menu.set_currently_syncing(syncing)

    def notify(self, evt: Event) -> None:
        title, msg, icon, duration = self._messages.get_msg(evt)
        self._show_msg(title, msg, icon, duration)

    def _show_msg(self, title: str, msg: str,
                  icon: QSystemTrayIcon.MessageIcon,
                  duration: int) -> None:
        duration = int(duration) * 1000
        self.showMessage(title, msg, icon, duration)



from textwrap import dedent

body_template = dedent("""
    <html>
      <head>
        <title>{title}</title>
        <meta name="viewport"
              content="width=device-width, initial-scale=1" />
        <script>
        const openShare = () => {{
          if(navigator.share) {{
            navigator.share({{
              title: '{title}',
              text: '{body_text}'
            }}).catch(console.error);
          }}
        }}

        document.addEventListener('DOMContentLoaded', function() {{
          const shareButton = document.getElementById('share-button');
          shareButton.style.display = (navigator.share) ? 'block' : 'none';
        }});
        </script>
        <style>
          :root {{ color-scheme: light dark; }}
          html {{ height: 100%; background-color: #212121; }}
          body {{
            height: calc(100% - 8rem);
            display: flex;
            flex-flow: column nowrap;
            align-items: flex-start;
            justify-content: stretch;
            margin: 2rem;
            padding: 2rem;
            border-radius: 1.5rem;
            box-shadow: 0px 2px 10px 2px black;
            background-color: light-dark(white, #212121);
            font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
            line-height: 1.5;
          }}
          header {{
            display: flex;
            flex-flow: row nowrap;
            justify-content: space-between;
            align-items: center;
            width: 100%;
          }}
          header > h2 {{ flex: 1; }}
          main {{ flex: 1; }}
          footer {{
            display: flex;
            flex-flow: column nowrap;
            align-self: center;
            align-items: center;
            font-size: 0.75rem;
            color: grey;
          }}
          footer > p {{ margin: 0.25rem; }}
          footer > a {{
            text-decoration: none;
            color: #a02c2c;
          }}
          ul#sharing {{
            display: flex;
            flex-flow: row nowrap;
            justify-content: center;
            margin: 0;
            padding: 0;
            grid-gap: 2rem;
          }}
          ul#sharing > li {{ list-style-type: none; }}
          ul#sharing svg {{
            fill: light-dark(black, white);
            cursor: pointer;
          }}
          #share-button {{ display: none; }}
        </style>
      </head>
      <body>
        <header>
          <h2>{title}</h2>
          <ul id="sharing">
            <li>
              <a href="mailto:?subject={title}&body={body_text}">
                <svg xmlns="http://www.w3.org/2000/svg"
                     viewBox="0 0 16 16" width="16" height="16">
                  <path d="M1.75 2h12.5c.966 0 1.75.784 1.75 1.75v8.5A1.75
                           1.75 0 0 1 14.25 14H1.75A1.75 1.75 0 0 1 0
                           12.25v-8.5C0 2.784.784 2 1.75 2ZM1.5 12.251c0
                           .138.112.25.25.25h12.5a.25.25 0 0 0
                           .25-.25V5.809L8.38 9.397a.75.75 0 0 1-.76 0L1.5
                           5.809v6.442Zm13-8.181v-.32a.25.25 0 0
                           0-.25-.25H1.75a.25.25 0 0
                           0-.25.25v.32L8 7.88Z"></path></svg>
              </a>
            </li>
            <li id="share-button" onClick="openShare()">
              <svg xmlns="http://www.w3.org/2000/svg"
                   viewBox="0 0 16 16" width="16" height="16">
                <path d="M15 3a3 3 0 0 1-5.175 2.066l-3.92 2.179a2.994 2.994
                         0 0 1 0 1.51l3.92 2.179a3 3 0 1 1-.73
                         1.31l-3.92-2.178a3 3 0 1 1 0-4.133l3.92-2.178A3
                         3 0 1 1 15 3Zm-1.5 10a1.5 1.5 0 1 0-3.001.001A1.5
                         1.5 0 0 0 13.5 13Zm-9-5a1.5 1.5 0 1 0-3.001.001A1.5
                         1.5 0 0 0 4.5 8Zm9-5a1.5 1.5 0 1 0-3.001.001A1.5
                         1.5 0 0 0 13.5 3Z"></path></svg>
            </li>
          </ul>
        </header>
        <main>{body_html}</main>
        <footer>
          <p>You may have to configure your browser to open
          attachments with the right application instead of downloading
          them again.</p>
          <p>Content downloaded by
          <a href="https://bbsync.app">Blackboard Sync</a></p>
        </footer>
      </body>
    </html>""").lstrip()


def create_body(title: str, body_html: str, body_text: str) -> str:
    return body_template.format(title=title,
                                body_html=body_html,
                                body_text=body_text)



from pathlib import Path
from requests import Response

from concurrent.futures import ThreadPoolExecutor


class BStream:
    """Base class for content that can be downloaded as a byte stream."""
    CHUNK_SIZE = 1024

    def write_base(self, path: Path, executor: ThreadPoolExecutor,
                   stream: Response) -> None:
        """Schedule the write operation."""

        def _write() -> None:
            with path.open("wb") as f:
                for chunk in stream.iter_content(chunk_size=self.CHUNK_SIZE):
                    f.write(chunk)

        executor.submit(_write)


class FStream:
    """Base class for content that can be written as text."""

    def write_base(self, path: Path, executor: ThreadPoolExecutor,
                   body: str) -> None:
        """Schedule the write operation."""

        def _write() -> None:
            with path.open('w', encoding='utf-8') as f:
                f.write(body)

        executor.submit(_write)



# Copyright (C) 2024, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import webbrowser
from threading import Timer
from requests.cookies import RequestsCookieJar

from PyQt6.QtCore import pyqtSlot, pyqtSignal, QObject, QUrl
from PyQt6.QtWidgets import QWidget, QPushButton, QLabel
from PyQt6.QtNetwork import QNetworkCookie
from PyQt6.QtWebEngineCore import QWebEnginePage, QWebEngineProfile
from PyQt6.QtWebEngineWidgets import QWebEngineView

from .assets import load_ui, get_theme_icon, AppIcon


WATCHDOG_DELAY = 30
PROFILE_STORAGE_NAME = "BlackboardLogin"


class LoginWebView(QWidget):
    """Login to a Blackboard instance using a browser."""

    class Signals(QObject):
        login_complete = pyqtSignal()

    def __init__(self, help_url: str) -> None:
        super().__init__()

        # Typing information
        self.web_view: QWebEngineView
        self.home_button: QPushButton
        self.back_button: QPushButton
        self.help_button: QPushButton
        self.status: QLabel

        self.watchdog: Timer | None

        self.start_url: str | None = None
        self.target_url: str | None = None

        self._cookie_jar = RequestsCookieJar()
        self.help_url = help_url

        self.signals = self.Signals()
        self._init_ui()

    def _init_ui(self) -> None:
        load_ui(self)

        self.home_button.setIcon(get_theme_icon(AppIcon.HOME))
        self.back_button.setIcon(get_theme_icon(AppIcon.BACK))
        self.help_button.setIcon(get_theme_icon(AppIcon.HELP))
        self.help_button.setVisible(False)

        # Since Qt6 a persistent profile must be created manually
        self.profile = QWebEngineProfile(PROFILE_STORAGE_NAME)
        self.page = QWebEnginePage(self.profile)
        self.web_view.setPage(self.page)

        self.init_signals()

    def init_signals(self) -> None:
        self.web_view.loadFinished.connect(self.slot_load_finished)
        self.home_button.clicked.connect(self.home)
        self.back_button.clicked.connect(self.web_view.back)
        self.help_button.clicked.connect(self.slot_help)
        self.profile.clearHttpCacheCompleted.connect(self.slot_cache_cleared)

        cookie_store = self.profile.cookieStore()
        if cookie_store is not None:
            cookie_store.cookieAdded.connect(self.slot_cookie_added)

    def load(self, start_url: str | None, target_url: str | None) -> None:
        self.start_url = start_url
        self.target_url = target_url

        # start watchdog
        self.watchdog = Timer(WATCHDOG_DELAY, self.show_help)
        self.watchdog.start()

        self.home()

    def home(self) -> None:
        self.web_view.load(QUrl.fromUserInput(self.start_url))

    @pyqtSlot()
    def slot_load_finished(self) -> None:
        """Check if we have reached the target url."""
        if self.target_url and self.url.startswith(self.target_url):
            if self.watchdog:
                self.watchdog.cancel()
            self.signals.login_complete.emit()

    @pyqtSlot(QNetworkCookie)
    def slot_cookie_added(self, cookie: QNetworkCookie) -> None:
        """Add the cookie to our own jar."""
        self._cookie_jar.set(
            cookie.name().data().decode(),
            cookie.value().data().decode(),
            domain=cookie.domain(),
            path=cookie.path(),
            secure=cookie.isSecure()
        )

    @pyqtSlot()
    def slot_cache_cleared(self) -> None:
        self.home()

    @pyqtSlot()
    def slot_help(self) -> None:
        webbrowser.open(self.help_url)

    def show_help(self) -> None:
        self.status.setText(self.tr(
            "Trouble logging in? Press the help button to let us know."
        ))
        self.help_button.setVisible(True)

    def restore(self) -> None:
        """Restore web view to original state."""
        self.clear_browser()

    def clear_browser(self) -> None:
        cookie_store = self.profile.cookieStore()
        if cookie_store is not None:
            cookie_store.deleteAllCookies()

        self.profile.clearHttpCache()
        self._cookie_jar = RequestsCookieJar()

    def shutdown(self) -> None:
        self.web_view.setPage(None)
        # Should destroy underlying Qt Objects
        del self.page
        del self.profile

        if self.watchdog:
            self.watchdog.cancel()

    @property
    def url(self) -> str:
        """URL of current website."""
        return self.web_view.url().toString()

    @property
    def cookies(self) -> RequestsCookieJar:
        """Contains session cookies of the current session."""
        return self._cookie_jar



"""BlackboardSync Configuration Manager Tests"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import tempfile
from pathlib import Path
from datetime import datetime
from unittest.mock import Mock, patch

import pytest
from hypothesis import given, assume
from hypothesis import strategies as st

from blackboard_sync.config import SyncConfig


def test_config_default_values():
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        s = SyncConfig(tmp_path)
        assert s.last_sync_time is None

@given(st.datetimes())
def test_config_last_sync_time(sync_time):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        config = SyncConfig(tmp_path)
        config.last_sync_time = sync_time
        assert config.last_sync_time == sync_time

        new_config = SyncConfig(tmp_path)
        assert new_config.last_sync_time == sync_time

